import sys
from unittest import mock

from grpc_router.server import main


@mock.patch('grpc_router.server.serve')
def test_server_entrypoint(serve_mock):
    args = ["server.py", "--hostname", "localhost", "--port", "9567", "--max-workers", "10"]
    with mock.patch.object(sys, 'argv', args):
        main()
    serve_mock.assert_called_once()
    config_options = serve_mock.call_args[0][0]
    assert config_options.hostname == "localhost"
    assert config_options.port == 9567
    assert config_options.max_workers == 10
    assert config_options.allow_global_region is True
    assert config_options.allow_cross_region_connectivity is True



import grpc
import pytest

from grpc_router.client.client import GRPCRouterClient


@pytest.mark.parametrize("service_id,host,port,region,expected_err_code,expected_err_str", [
    ["", "myhost", 1234, "NYC", grpc.StatusCode.INVALID_ARGUMENT, "The service_id cannot be empty."],
    ["my,svc", "", 1234, "NYC", grpc.StatusCode.INVALID_ARGUMENT, "The host cannot be empty."],
    ["my,svc", "myhost", -10, "NYC", grpc.StatusCode.INVALID_ARGUMENT, "The port cannot be negative or zero."],
    ["my,svc", "myhost", 0, "NYC", grpc.StatusCode.INVALID_ARGUMENT, "The port cannot be negative or zero."],
])
def test_register_service_validation_errors(service_id, host, port, region, expected_err_code, expected_err_str, grpc_router_server):
    client = GRPCRouterClient("localhost", 7654)
    with pytest.raises(grpc.RpcError) as exc:
        client.register_service(
            service_id=service_id,
            host=host,
            port=port,
            region=region
        )
    assert exc.value.code() == expected_err_code
    assert exc.value.details() == expected_err_str


@pytest.mark.parametrize("service_id,host,port,region,expected_err_code,expected_err_str", [
    ["my,svc", "myhost", 1234, "", grpc.StatusCode.INVALID_ARGUMENT, "The region cannot be empty in this current configuration."],
])
def test_register_service_validation_errors_no_global_region(service_id, host, port, region, expected_err_code, expected_err_str, grpc_router_server_no_allow_global_region):
    client = GRPCRouterClient("localhost", 7653)
    with pytest.raises(grpc.RpcError) as exc:
        client.register_service(
            service_id=service_id,
            host=host,
            port=port,
            region=region
        )
    assert exc.value.code() == expected_err_code
    assert exc.value.details() == expected_err_str


@pytest.mark.parametrize("service_id,service_token,expected_err_code,expected_err_str", [
    ["", "token", grpc.StatusCode.INVALID_ARGUMENT, "The service_id cannot be empty."],
    ["my,svc", "", grpc.StatusCode.INVALID_ARGUMENT, "The service_context cannot be empty."],
])
def test_deregister_service_validation_errors(service_id, service_token, expected_err_code, expected_err_str, grpc_router_server):
    client = GRPCRouterClient("localhost", 7654)
    with pytest.raises(grpc.RpcError) as exc:
        client.deregister_service(
            service_id=service_id,
            service_token=service_token
        )
    assert exc.value.code() == expected_err_code
    assert exc.value.details() == expected_err_str


@pytest.mark.parametrize("service_id,expected_err_code,expected_err_str", [
    ["", grpc.StatusCode.INVALID_ARGUMENT, "The service_id cannot be empty."],
])
def test_get_registered_service_validation_errors(service_id, expected_err_code, expected_err_str, grpc_router_server):
    client = GRPCRouterClient("localhost", 7654)
    with pytest.raises(grpc.RpcError) as exc:
        client.get_service(
            service_id=service_id
        )
    assert exc.value.code() == expected_err_code
    assert exc.value.details() == expected_err_str



from grpc_router.client.client import GRPCRouterClient
from grpc_reflection.v1alpha import reflection_pb2
from grpc_reflection.v1alpha import reflection_pb2_grpc


def test_reflection(grpc_router_server):
    client = GRPCRouterClient("localhost", 7654)

    stub = reflection_pb2_grpc.ServerReflectionStub(client.channel)
    request = reflection_pb2.ServerReflectionRequest()
    request.host = ""
    request.list_services = ""
    res = stub.ServerReflectionInfo((r for r in (request,)))
    responses = list(res)
    assert len(responses) == 1
    response = responses[0]
    services = [r.name for r in response.list_services_response.service]
    assert services == [
        'grpc.reflection.v1alpha.ServerReflection',
        'grpcrouter.GRPCRouterService'
    ]



import argparse
from concurrent.futures import ThreadPoolExecutor
import grpc
from grpc_health.v1 import health
from grpc_health.v1 import health_pb2
from grpc_health.v1 import health_pb2_grpc
from grpc_reflection.v1alpha import reflection

from grpc_router.stubs.grpc_router_service_pb2_grpc import add_GRPCRouterServiceServicer_to_server, GRPCRouterServiceServicer
from grpc_router.stubs.grpc_router_service_pb2 import (
    ServiceRegistrationResponse,
    ServiceDeregistrationResponse,
    GetRegisteredServiceResponse,
    DESCRIPTOR,
)

from grpc_router.core.models import ConfigOptions
from grpc_router.core.register import ServiceRegister


class GRPCRouterServer(GRPCRouterServiceServicer):

    def __init__(self, config: ConfigOptions):
        self._config = config
        self._register = ServiceRegister(config)

    def _validate_RegisterService(self, request, context):
        service_id = request.service_id
        if not service_id:
            context.abort(
                grpc.StatusCode.INVALID_ARGUMENT,
                "The service_id cannot be empty."
            )
        host = request.endpoint.host
        if not host:
            context.abort(
                grpc.StatusCode.INVALID_ARGUMENT,
                "The host cannot be empty."
            )

        port = request.endpoint.port
        if port <= 0:
            context.abort(
                grpc.StatusCode.INVALID_ARGUMENT,
                "The port cannot be negative or zero."
            )
        region = request.metadata.region
        if not region and not self._config.allow_global_region:
            context.abort(
                grpc.StatusCode.INVALID_ARGUMENT,
                "The region cannot be empty in this current configuration."
            )

    def RegisterService(self, request, context) -> ServiceRegistrationResponse:
        self._validate_RegisterService(request, context)
        service_token, error = self._register.register_service(
            service_id=request.service_id,
            host=request.endpoint.host,
            port=request.endpoint.port,
            region=request.metadata.region,
            slots=request.metadata.slots,
        )
        return ServiceRegistrationResponse(
            service_token=service_token
        )

    def _validate_DeregisterService(self, request, context) -> None:
        service_id = request.service_id
        if not service_id:
            context.abort(
                grpc.StatusCode.INVALID_ARGUMENT,
                "The service_id cannot be empty."
            )
        service_token = request.service_token
        if not service_token:
            context.abort(
                grpc.StatusCode.INVALID_ARGUMENT,
                "The service_context cannot be empty."
            )

    def DeregisterService(self, request, context) -> ServiceDeregistrationResponse:
        self._validate_DeregisterService(request, context)
        self._register.deregister_service(
            service_id=request.service_id,
            service_token=request.service_token
        )
        return ServiceDeregistrationResponse()

    def _validate_GetRegisteredService(self, request, context) -> None:
        service_id = request.service_id
        if not service_id:
            context.abort(
                grpc.StatusCode.INVALID_ARGUMENT,
                "The service_id cannot be empty."
            )

    def GetRegisteredService(self, request, context) -> GetRegisteredServiceResponse:
        self._validate_GetRegisteredService(request, context)
        service = self._register.get_service(
            service_id=request.service_id,
            region=request.hints.region
        )
        if service is None:
            context.abort(
                grpc.StatusCode.NOT_FOUND,
                "The service_id has no registered instances."
            )
        assert service is not None
        response = GetRegisteredServiceResponse()
        response.service_id = service.service_id
        response.endpoint.host = service.host
        response.endpoint.port = service.port
        return response


def enable_reflection_service(server: grpc.Server) -> None:
    SERVICE_NAMES = (
        DESCRIPTOR.services_by_name["GRPCRouterService"].full_name,
        reflection.SERVICE_NAME,
    )
    reflection.enable_server_reflection(SERVICE_NAMES, server)


def enable_health_service(server: grpc.Server) -> None:
    health_servicer = health.HealthServicer(
        experimental_non_blocking=True,
        experimental_thread_pool=ThreadPoolExecutor(max_workers=5),
    )
    health_pb2_grpc.add_HealthServicer_to_server(health_servicer, server)
    health_servicer.set(
        DESCRIPTOR.services_by_name["GRPCRouterService"].full_name,
        health_pb2.HealthCheckResponse.SERVING
    )


def serve(config: ConfigOptions) -> None:
    server = grpc.server(ThreadPoolExecutor(max_workers=config.max_workers))
    add_GRPCRouterServiceServicer_to_server(GRPCRouterServer(config), server)
    enable_reflection_service(server)
    enable_health_service(server)
    server.add_insecure_port(f"{config.hostname}:{config.port}")
    server.start()
    server.wait_for_termination()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-n', '--hostname', dest='hostname', default='[::]',
                        help='Hostname to bind the service to')
    parser.add_argument('-p', '--port', dest='port', default=50034, type=int,
                        help='Port to bind this service to')
    parser.add_argument('-w', '--max-workers', dest='max_workers',
                        type=int, default=10,
                        help='Maximum concurrent workers to handle requests.')

    args = parser.parse_args()
    config = ConfigOptions(
        hostname=args.hostname,
        port=args.port,
        max_workers=args.max_workers,
    )
    serve(config)


if __name__ == "__main__":
    main()



from grpc_router.client.client import GRPCRouterClient
from grpc_health.v1 import health_pb2
from grpc_health.v1 import health_pb2_grpc


def test_health_check(grpc_router_server):
    client = GRPCRouterClient("localhost", 7654)

    stub = health_pb2_grpc.HealthStub(client.channel)
    res = stub.Check(
        health_pb2.HealthCheckRequest(service="")
    )
    assert res.status == health_pb2.HealthCheckResponse.SERVING
    res = stub.Check(
        health_pb2.HealthCheckRequest(service="grpcrouter.GRPCRouterService")
    )
    assert res.status == health_pb2.HealthCheckResponse.SERVING



import grpc
import pytest

from grpc_router.client.client import GRPCRouterClient


def register_svc(client: GRPCRouterClient, service_id: str, region: str, instance: int, port: int) -> str:
    token = client.register_service(
        service_id=service_id,
        host=f"{region}{instance}.mydomain.com",
        port=port,
        region=region
    )
    assert token is not None
    return token


@pytest.mark.parametrize('region,expected_host,expected_port', [
    ["USWest", "USWest1.mydomain.com", 9002],
    ["USWest", "USWest2.mydomain.com", 9003],
    ["USWest", "USWest3.mydomain.com", 9004],
    ["USEast", "USEast1.mydomain.com", 9000],
    ["USEast", "USEast2.mydomain.com", 9001],
    ["USWest", "USWest1.mydomain.com", 9002],
    ["USEast", "USEast1.mydomain.com", 9000],
    ["", "1.mydomain.com", 9005],
    ["", "2.mydomain.com", 9006],
    ["", "1.mydomain.com", 9005],
    ["UNKNOWN", "2.mydomain.com", 9006],
    ["UNKNOWN2", "1.mydomain.com", 9005],
])
def test_region_stickiness(region, expected_host, expected_port, grpc_router_server):
    # the instance in the same region is preferred over other regions
    service_id = "my.region.stickiness.test.service"

    client = GRPCRouterClient("localhost", 7654)

    tokens = [
        register_svc(client, service_id, "USEast", 1, 9000),
        register_svc(client, service_id, "USEast", 2, 9001),
        register_svc(client, service_id, "USWest", 1, 9002),
        register_svc(client, service_id, "USWest", 2, 9003),
        register_svc(client, service_id, "USWest", 3, 9004),
        register_svc(client, service_id, "", 1, 9005),
        register_svc(client, service_id, "", 2, 9006),
    ]

    try:
        host, port = client.get_service(service_id, region=region)
        assert host == expected_host
        assert port == expected_port
    finally:
        for token in tokens:
            client.deregister_service(
                service_id=service_id,
                service_token=token
            )


@pytest.mark.parametrize('region,expected_host,expected_port', [
    ["USWest", "USWest1.mydomain.com", 9002],
    ["USWest", "USWest2.mydomain.com", 9003],
    ["USWest", "USWest3.mydomain.com", 9004],
    ["USEast", "USEast1.mydomain.com", 9000],
    ["", "USEast1.mydomain.com", 9000],
    ["USEast", "USEast2.mydomain.com", 9001],
    ["USCentral", "USCentral1.mydomain.com", 9005],
    ["", "USEast2.mydomain.com", 9001],
    ["", "USWest1.mydomain.com", 9002],
    ["USWest", "USWest1.mydomain.com", 9002],
    ["USEast", "USEast1.mydomain.com", 9000],
    ["UNKNOWN", "USWest2.mydomain.com", 9003],
    ["UNKNOWN2", "USWest3.mydomain.com", 9004],
    ["USCentral", "USCentral2.mydomain.com", 9006],
    ["UNKNOWN", "USCentral1.mydomain.com", 9005],
    ["UNKNOWN", "USCentral2.mydomain.com", 9006],
    ["NOREGION", "USEast1.mydomain.com", 9000],
])
def test_region_no_global_region(region, expected_host, expected_port, grpc_router_server):
    # the instance in the same region is preferred over other regions
    service_id = "my.region.noglobal.test.service"

    client = GRPCRouterClient("localhost", 7654)

    tokens = []

    tokens = [
        register_svc(client, service_id, "USEast", 1, 9000),
        register_svc(client, service_id, "USEast", 2, 9001),
        register_svc(client, service_id, "USWest", 1, 9002),
        register_svc(client, service_id, "USWest", 2, 9003),
        register_svc(client, service_id, "USWest", 3, 9004),
        register_svc(client, service_id, "USCentral", 1, 9005),
        register_svc(client, service_id, "USCentral", 2, 9006),
    ]

    try:
        host, port = client.get_service(service_id, region=region)
        assert host == expected_host
        assert port == expected_port
    finally:
        for token in tokens:
            client.deregister_service(
                service_id=service_id,
                service_token=token
            )


def test_region_cross_region(grpc_router_server):
    service_id = "my.region.cross_region.test.service"

    client = GRPCRouterClient("localhost", 7654)

    tokens = []

    tokens = [
        register_svc(client, service_id, "USEast", 1, 9000),
        register_svc(client, service_id, "USEast", 2, 9001),
        register_svc(client, service_id, "USWest", 1, 9002),
    ]
    try:
        host, port = client.get_service(service_id, region="USMidWest")
        assert host == "USEast1.mydomain.com"
        assert port == 9000
    finally:
        for token in tokens:
            client.deregister_service(
                service_id=service_id,
                service_token=token
            )


def test_region_no_cross_region(grpc_router_server_no_global_region_no_cross_region):
    service_id = "my.region.no_cross_region.test.service"

    client = GRPCRouterClient("localhost", 7652)

    tokens = []

    tokens = [
        register_svc(client, service_id, "USEast", 1, 9000),
        register_svc(client, service_id, "USEast", 2, 9001),
        register_svc(client, service_id, "USWest", 1, 9002),
    ]
    try:
        with pytest.raises(grpc.RpcError) as exc:
            client.get_service(service_id, region="USMidWest")
        assert exc.value.code() == grpc.StatusCode.NOT_FOUND
        assert exc.value.details() == "The service_id has no registered instances."
    finally:
        for token in tokens:
            client.deregister_service(
                service_id=service_id,
                service_token=token
            )



import grpc

from grpc_router.client.client import GRPCRouterClient
import pytest


def test_connectivity(grpc_router_server):
    service_id = "my.test.service"
    client = GRPCRouterClient("localhost", 7654)

    with pytest.raises(grpc.RpcError) as exc:
        client.get_service(service_id)
    assert exc.value.code() == grpc.StatusCode.NOT_FOUND
    assert exc.value.details() == "The service_id has no registered instances."

    token = client.register_service(
        service_id=service_id,
        host="myhost.mydomain.com",
        port=9998
    )
    assert token is not None

    host, port = client.get_service(service_id)
    assert host == "myhost.mydomain.com"
    assert port == 9998

    client.deregister_service(
        service_id=service_id,
        service_token=token
    )

    with pytest.raises(grpc.RpcError) as exc:
        client.get_service(service_id)
    assert exc.value.code() == grpc.StatusCode.NOT_FOUND
    assert exc.value.details() == "The service_id has no registered instances."


def test_multiple_services_round_robin(grpc_router_server):
    service_id = "my.own.test.service"

    client = GRPCRouterClient("localhost", 7654)

    token1 = client.register_service(
        service_id=service_id,
        host="myhost1.mydomain.com",
        port=9990
    )
    token2 = client.register_service(
        service_id=service_id,
        host="myhost2.mydomain.com",
        port=9991
    )
    token3 = client.register_service(
        service_id=service_id,
        host="myhost3.mydomain.com",
        port=9992
    )

    host, port = client.get_service(service_id)
    assert host == "myhost1.mydomain.com"
    assert port == 9990
    host, port = client.get_service(service_id)
    assert host == "myhost2.mydomain.com"
    assert port == 9991
    host, port = client.get_service(service_id)
    assert host == "myhost3.mydomain.com"
    assert port == 9992
    host, port = client.get_service(service_id)
    assert host == "myhost1.mydomain.com"
    assert port == 9990

    client.deregister_service(
        service_id=service_id,
        service_token=token1
    )
    client.deregister_service(
        service_id=service_id,
        service_token=token2
    )
    client.deregister_service(
        service_id=service_id,
        service_token=token3
    )
